/*
 *
 * ------------------------------------------------------------------------------------------------------------
 * Copyright (c) 2022 InvenSense, Inc All rights reserved.
 *
 * This software, related documentation and any modifications thereto (collectively "Software") is subject
 * to InvenSense, Inc and its licencors' intellectual property rights under U.S. and international copyright
 * and other intellectual property rights laws.
 *
 * InvenSense, Inc and its licencors retain all intellectual property and proprietary rights in and to the Software
 * and any use, reproduction, disclosure or distribution of the Software without an express license agreement
 * from InvenSense, Inc is strictly prohibited.
 *
 * EXCEPT AS OTHERWISE PROVIDED IN A LICENSE AGREEMENT BETWEEN THE PARTIES, THE SOFTWARE IS
 * PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.
 * EXCEPT AS OTHERWISE PROVIDED IN A LICENSE AGREEMENT BETWEEN THE PARTIES, IN NO EVENT SHALL
 * InvenSense, Inc BE LIABLE FOR ANY DIRECT, SPECIAL, INDIRECT, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, OR ANY
 * DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTUOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
 * OF THE SOFTWARE.
 *
 * ------------------------------------------------------------------------------------------------------------
 */
#include <zephyr/kernel.h>
#include <zephyr/device.h>
#include <zephyr/drivers/i2c.h>
//#include <nrfx_gpiote.h>
#include <zephyr/drivers/gpio.h>

#include "ICM42670P.h"

static int ii2c_write(struct inv_imu_serif *serif, uint8_t reg, const uint8_t *wbuffer, uint32_t wlen);
static int ii2c_read(struct inv_imu_serif *serif, uint8_t reg, uint8_t *rbuffer, uint32_t rlen);
static void event_cb(inv_imu_sensor_event_t *event);

// i2c and SPI interfaces are used from C driver callbacks, without any knowledge of the object
// As they are declared as static, they will be overriden each time a new ICM42670P object is created
// i2c
uint8_t i2c_address = 0;

#define ICM42670P_I2C_SPEED 1000000
#define ICM42670P_I2C_ADDRESS 0x68

const struct device *i2c_dev = DEVICE_DT_GET(DT_NODELABEL(i2c0));


static const char* APEX_ACTIVITY[3] = {"IDLE","WALK","RUN"};
ICM42670P_irq_handler my_handler;

void check_i2c(uint16_t addr)
{
    struct i2c_msg msgs[1];
    uint8_t data;

    msgs[0].buf = &data;
    msgs[0].len = 1;
    msgs[0].flags = I2C_MSG_READ | I2C_MSG_STOP;

    int ret = i2c_transfer(i2c_dev, &msgs[0], 1, addr);
    if (ret == 0)
    {
        //printk("Found device at address 0x%02X\n", addr);
    }
    else if (ret == -ENODEV)
    {
        //  continue;  // No device at this address, continue scanning
    }
    else
    {
        //printk("Error scanning device at address 0x%02X\n", addr);
    }
}

// This is used by the event callback (not object aware), declared static
static inv_imu_sensor_event_t *event;

// This is used by the getDataFromFifo callback (not object aware), declared static
static struct inv_imu_device *icm_driver_ptr = NULL;

struct gpio_callback int1_gyro_cb_data;
struct gpio_callback int2_gyro_cb_data;

static const struct gpio_dt_spec int1_gyro = GPIO_DT_SPEC_GET(DT_NODELABEL(gyro_int1), gpios);
static const struct gpio_dt_spec int2_gyro = GPIO_DT_SPEC_GET(DT_NODELABEL(gyro_int2), gpios);

struct inv_imu_serif icm_serif;

// ICM42670P constructor for I2c interface
ICM42670P::ICM42670P(bool lsb)
{
    i2c_address = ICM42670P_I2C_ADDRESS | (lsb ? 0x1 : 0);
    my_handler = NULL;
}

/* starts communication with the ICM42670P */
int ICM42670P::begin()
{
    
    int rc = 0;
    uint8_t who_am_i;

    if (!device_is_ready(i2c_dev))
    {
        //printk("I2C: Device is not ready.\n");
        return -1;
    }

    //k_msleep(500);

    int result = i2c_configure(i2c_dev, I2C_SPEED_SET(ICM42670P_I2C_SPEED));

    check_i2c(ICM42670P_I2C_ADDRESS);

    icm_serif.serif_type = UI_I2C;
    icm_serif.read_reg = ii2c_read;
    icm_serif.write_reg = ii2c_write;

    /// Initialize serial interface between MCU and Icm43xxx
    icm_serif.context = (void*)this;
    icm_serif.max_read = 2048;  // maximum number of bytes allowed per serial read
    icm_serif.max_write = 2048; // maximum number of bytes allowed per serial write
    
    icm_driver.sensor_event_cb = event_cb;

    rc = inv_imu_init(&icm_driver, &icm_serif, NULL);
    
    
    gpio_pin_configure_dt(&int1_gyro, GPIO_INPUT);
    //gpio_pin_interrupt_configure_dt(&int1_gyro, GPIO_INT_EDGE_TO_INACTIVE);
    gpio_pin_interrupt_configure_dt(&int1_gyro, GPIO_INT_TRIG_LOW | GPIO_INT_MODE_EDGE);
    //gpio_pin_configure_dt(&int2_gyro, GPIO_INPUT);
    //gpio_pin_interrupt_configure_dt(&int2_gyro, GPIO_INT_EDGE_TO_ACTIVE);
    if (rc != INV_ERROR_SUCCESS)
    {
        // printk("inv_imu_init error\n");
        return rc;
    }

    

    /// Check WHOAMI
    rc = inv_imu_get_who_am_i(&icm_driver, &who_am_i);
    if (rc != 0)
    {
        return -2;
    }
    if (who_am_i != INV_IMU_WHOAMI) {
      return -3;
    }

    // printk("successful init %d\n", who_am_i);

    // successful init, return 0
    return 0;
}

int ICM42670P::reinit()
{
    struct inv_imu_serif icm_serif;
    icm_serif.serif_type = UI_I2C;
    icm_serif.read_reg = ii2c_read;
    icm_serif.write_reg = ii2c_write;

    /// Initialize serial interface between MCU and Icm43xxx
    icm_serif.context = 0;      // no need
    icm_serif.max_read = 2048;  // maximum number of bytes allowed per serial read
    icm_serif.max_write = 2048; // maximum number of bytes allowed per serial write

    icm_driver.sensor_event_cb = event_cb;

    int rc = inv_imu_init(&icm_driver, &icm_serif, NULL);
    
    //gpio_pin_configure_dt(&int1_gyro, GPIO_INPUT);
    //gpio_pin_interrupt_configure_dt(&int1_gyro, GPIO_INT_EDGE_TO_INACTIVE);
    //gpio_pin_configure_dt(&int2_gyro, GPIO_INPUT);
    //gpio_pin_interrupt_configure_dt(&int2_gyro, GPIO_INT_EDGE_TO_ACTIVE);
    if (rc != INV_ERROR_SUCCESS)
    {
        // printk("inv_imu_init error\n");
        return rc;
    }
    return 0;
}

int ICM42670P::startAccel(uint16_t odr, uint16_t fsr)
{
    int rc = 0;
    rc |= inv_imu_set_accel_fsr(&icm_driver, accel_fsr_g_to_param(fsr));
    rc |= inv_imu_set_accel_frequency(&icm_driver, accel_freq_to_param(odr));
    rc |= inv_imu_enable_accel_low_noise_mode(&icm_driver);
    return rc;
}

int ICM42670P::startGyro(uint16_t odr, uint16_t fsr)
{
    int rc = 0;
    rc |= inv_imu_set_gyro_fsr(&icm_driver, gyro_fsr_dps_to_param(fsr));
    rc |= inv_imu_set_gyro_frequency(&icm_driver, gyro_freq_to_param(odr));
    rc |= inv_imu_enable_gyro_low_noise_mode(&icm_driver);
    return rc;
}

int ICM42670P::getDataFromRegisters(inv_imu_sensor_event_t *evt)
{
    if (evt != NULL)
    {
        // Set event buffer to be used by the callback
        //event = evt;
        return inv_imu_get_data_from_registers(&icm_driver, evt);
    }
    else
    {
        return -1;
    }
}


void button_callback(const struct device *dev, struct gpio_callback *cb, uint32_t pins)
{
    if (my_handler == NULL)
    {
        return;
    }

    my_handler();
}

int ICM42670P::initApex(uint8_t intpin, ICM42670P_irq_handler handler)
{
  int                       rc = 0;
  inv_imu_apex_parameters_t apex_inputs;
  inv_imu_interrupt_parameter_t config_int = { (inv_imu_interrupt_value)0 };

  /* Disabling FIFO usage to optimize power consumption */
  rc |= inv_imu_configure_fifo(&icm_driver, INV_IMU_FIFO_DISABLED);

  /* Disable FIFO threshold, DRDY and WOM interrupts on INT1 */
  rc |= inv_imu_get_config_int1(&icm_driver, &config_int);
  config_int.INV_FIFO_THS = INV_IMU_DISABLE;
  config_int.INV_UI_DRDY  = INV_IMU_DISABLE;
  config_int.INV_WOM_X    = INV_IMU_DISABLE;
  config_int.INV_WOM_Y    = INV_IMU_DISABLE;
  config_int.INV_WOM_Z    = INV_IMU_DISABLE;
  rc |= inv_imu_set_config_int1(&icm_driver, &config_int);

  /* Enable accel in LP mode */
  rc |= inv_imu_enable_accel_low_power_mode(&icm_driver);

  /* Disable Pedometer before configuring it */
  rc |= inv_imu_apex_disable_pedometer(&icm_driver);
  rc |= inv_imu_apex_disable_tilt(&icm_driver);

  rc |= inv_imu_set_accel_frequency(&icm_driver, ACCEL_CONFIG0_ODR_50_HZ);
  rc |= inv_imu_apex_set_frequency(&icm_driver, APEX_CONFIG1_DMP_ODR_50Hz);

  /* Set APEX parameters */
  rc |= inv_imu_apex_init_parameters_struct(&icm_driver, &apex_inputs);
  apex_inputs.power_save =APEX_CONFIG0_DMP_POWER_SAVE_DIS;
  rc |= inv_imu_apex_configure_parameters(&icm_driver, &apex_inputs);

  //pinMode(intpin,INPUT);
  //attachInterrupt(intpin,handler,RISING);

  return rc;
}


int ICM42670P::startTiltDetection(uint8_t intpin, ICM42670P_irq_handler handler)
{
  int rc = 0;
  rc |= initApex(intpin,handler);
  rc |= inv_imu_apex_enable_tilt(&icm_driver);

  return rc;
}

int ICM42670P::startPedometer(uint8_t intpin, ICM42670P_irq_handler handler)
{
  int rc = 0;
  step_cnt_ovflw = 0;
  rc |= initApex(intpin,handler);
  rc |= inv_imu_apex_enable_pedometer(&icm_driver);

  return rc;
}

int ICM42670P::getPedometer(uint32_t& step_count, float& step_cadence, const char*& activity)
{
  int rc = 0;
  uint8_t  int_status3;
  
  /* Read APEX interrupt status */
  rc |= inv_imu_read_reg(&icm_driver, INT_STATUS3, 1, &int_status3);

  if (int_status3 & INT_STATUS3_STEP_CNT_OVF_INT_MASK)
    step_cnt_ovflw++;

  if (int_status3 & (INT_STATUS3_STEP_DET_INT_MASK)) {
    inv_imu_apex_step_activity_t apex_data0;
    float nb_samples           = 0;

    rc |= inv_imu_apex_get_data_activity(&icm_driver, &apex_data0);
    // to do: detect step counter overflow?
    step_count = apex_data0.step_cnt + step_cnt_ovflw*(uint32_t)UINT16_MAX;	
    /* Converting u6.2 to float */
    nb_samples = (apex_data0.step_cadence >> 2) +
        (float)(apex_data0.step_cadence & 0x03) * 0.25f;
    if(nb_samples != 0)
    {
      step_cadence = (float)50 / nb_samples;
    } else {
      step_cadence = 0;
    }
    activity = APEX_ACTIVITY[apex_data0.activity_class];
  } else {
    return -11;  
  }

  return rc;
}

int ICM42670P::enableFifoInterrupt(uint8_t intpin, ICM42670P_irq_handler handler, uint8_t fifo_watermark)
{
    int rc = 0;
    uint8_t data;

    if (handler == NULL)
    {
        return -1;
    }

    my_handler = handler;

    

    gpio_pin_configure_dt(&int1_gyro, GPIO_INPUT);
    gpio_pin_interrupt_configure_dt(&int1_gyro, GPIO_INT_EDGE_TO_ACTIVE);

    gpio_init_callback(&int1_gyro_cb_data, button_callback, BIT(int1_gyro.pin));
    gpio_add_callback(int1_gyro.port, &int1_gyro_cb_data);

    // pinMode(intpin,INPUT);
    // attachInterrupt(intpin,handler,RISING);

    rc |= inv_imu_configure_fifo(&icm_driver, INV_IMU_FIFO_ENABLED);
    rc |= inv_imu_write_reg(&icm_driver, FIFO_CONFIG2, 1, &fifo_watermark);
    // Set fifo_wm_int_w generating condition : fifo_wm_int_w generated when counter == threshold
    rc |= inv_imu_read_reg(&icm_driver, FIFO_CONFIG5_MREG1, 1, &data);
    data &= (uint8_t)~FIFO_CONFIG5_WM_GT_TH_EN;
    rc |= inv_imu_write_reg(&icm_driver, FIFO_CONFIG5_MREG1, 1, &data);
    // Disable APEX to use 2.25kB of fifo for raw data
    data = SENSOR_CONFIG3_APEX_DISABLE_MASK;
    rc |= inv_imu_write_reg(&icm_driver, SENSOR_CONFIG3_MREG1, 1, &data);

    return rc;
}

int ICM42670P::configureEnableWOM(ICM42670P_irq_handler handler,uint8_t wom_x,uint8_t wom_y, uint8_t wom_z, uint8_t freq,  bool dual_int)
{
    int rc = 0;
    uint8_t data;

    if (handler == NULL)
    {
        return -1;
    }

    my_handler = handler;    

    //gpio_pin_configure_dt(&int1_gyro, GPIO_INPUT );
    //gpio_pin_interrupt_configure_dt(&int1_gyro, GPIO_INT_EDGE_TO_INACTIVE);

    gpio_init_callback(&int1_gyro_cb_data, button_callback, BIT(int1_gyro.pin));
    gpio_add_callback(int1_gyro.port, &int1_gyro_cb_data);

    if(dual_int) {
        //gpio_pin_configure_dt(&int2_gyro, GPIO_INPUT );
        //gpio_pin_interrupt_configure_dt(&int2_gyro, GPIO_INT_EDGE_TO_INACTIVE);

        gpio_init_callback(&int2_gyro_cb_data, button_callback, BIT(int2_gyro.pin));
        gpio_add_callback(int2_gyro.port, &int2_gyro_cb_data);
    }


    rc |= inv_imu_configure_fifo(&icm_driver, INV_IMU_FIFO_DISABLED);

    rc|=inv_imu_set_accel_fsr(&icm_driver,ACCEL_CONFIG0_FS_SEL_2g);    
    rc|=inv_imu_set_accel_frequency(&icm_driver,(ACCEL_CONFIG0_ODR_t)freq);
    if(freq>=7) {
        rc|=inv_imu_enable_accel_low_power_mode(&icm_driver);
    } else {
        rc|=inv_imu_enable_accel_low_noise_mode(&icm_driver);
    }    
    rc|=inv_imu_disable_gyro(&icm_driver);

    inv_imu_interrupt_parameter_t imu_int, imu_int2;
    //rc|=inv_imu_get_config_int1(&icm_driver,&imu_int);
    
    imu_int.INV_WOM_X=INV_IMU_ENABLE;
    imu_int.INV_WOM_Y=INV_IMU_ENABLE;
    imu_int.INV_WOM_Z=INV_IMU_ENABLE; 
       
    rc|=inv_imu_set_config_int1(&icm_driver,&imu_int);
    
        
    inv_imu_interrupt_config_t imu_int_config;
    imu_int_config.INV_INT2_MODE=0;
    imu_int_config.INV_INT2_DRIVE_CIRCUIT=0;
    imu_int_config.INV_INT2_POLARITY=0;
    imu_int_config.INV_INT1_MODE=0;
    imu_int_config.INV_INT1_DRIVE_CIRCUIT=1;
    imu_int_config.INV_INT1_POLARITY=0;
    rc|=inv_imu_set_int1_int2_config(&icm_driver,&imu_int_config);
    rc|=inv_imu_set_accel_lp_avg(&icm_driver,ACCEL_CONFIG1_ACCEL_FILT_AVG_2);
    rc|=inv_imu_configure_wom(&icm_driver,wom_x,wom_y,wom_z,WOM_CONFIG_WOM_INT_MODE_ORED,WOM_CONFIG_WOM_INT_DUR_1_SMPL);
    rc|=inv_imu_enable_wom(&icm_driver);
    if(dual_int) {
        imu_int2.INV_WOM_X=INV_IMU_ENABLE;
        imu_int2.INV_WOM_Y=INV_IMU_ENABLE;
        imu_int2.INV_WOM_Z=INV_IMU_ENABLE; 
        //step_cnt_ovflw = 0;
        //rc |= initApex(2,handler);
        //rc |= inv_imu_apex_enable_pedometer(&icm_driver);
        //rc |= inv_imu_apex_enable_tilt(&icm_driver);
        
        //imu_int2.INV_STEP_DET=INV_IMU_ENABLE;
        //imu_int2.INV_TILT_DET=INV_IMU_ENABLE;
        rc|=inv_imu_set_config_int2(&icm_driver,&imu_int2);
    }

    return rc;
}

int ICM42670P::getDataFromFifo(ICM42670P_sensor_event_cb event_cb)
{
    if (event_cb == NULL)
    {
        return -1;
    }
    icm_driver.sensor_event_cb = event_cb;
    return inv_imu_get_data_from_fifo(&icm_driver);
}

bool ICM42670P::isAccelDataValid(inv_imu_sensor_event_t *evt)
{
    return (evt->sensor_mask & (1 << INV_SENSOR_ACCEL));
}

bool ICM42670P::isGyroDataValid(inv_imu_sensor_event_t *evt)
{
    return (evt->sensor_mask & (1 << INV_SENSOR_GYRO));
}

static int ii2c_write(struct inv_imu_serif *serif, uint8_t reg, const uint8_t *wbuffer, uint32_t wlen)
{
    for (uint32_t i = 0; i < wlen; ++i)
    {
        if (i2c_reg_write_byte(i2c_dev, ICM42670P_I2C_ADDRESS, reg + i, wbuffer[i]))
        {
            // printk("Failed to write register\n");
            return -1;
        }
    }

    return 0;
}

static int ii2c_read(struct inv_imu_serif *serif, uint8_t reg, uint8_t *rbuffer, uint32_t rlen)
{
    for (uint32_t i = 0; i < rlen; ++i)
    {
        if (i2c_reg_read_byte(i2c_dev, ICM42670P_I2C_ADDRESS, reg + i, &rbuffer[i]))
        {
            // printk("Failed to read register\n");
            return -1;
        }
    }

    return 0;
}

ACCEL_CONFIG0_FS_SEL_t ICM42670P::accel_fsr_g_to_param(uint16_t accel_fsr_g)
{
    ACCEL_CONFIG0_FS_SEL_t ret = ACCEL_CONFIG0_FS_SEL_16g;

    switch (accel_fsr_g)
    {
    case 2:
        ret = ACCEL_CONFIG0_FS_SEL_2g;
        break;
    case 4:
        ret = ACCEL_CONFIG0_FS_SEL_4g;
        break;
    case 8:
        ret = ACCEL_CONFIG0_FS_SEL_8g;
        break;
    case 16:
        ret = ACCEL_CONFIG0_FS_SEL_16g;
        break;
    default:
        /* Unknown accel FSR. Set to default 16G */
        break;
    }
    return ret;
}

GYRO_CONFIG0_FS_SEL_t ICM42670P::gyro_fsr_dps_to_param(uint16_t gyro_fsr_dps)
{
    GYRO_CONFIG0_FS_SEL_t ret = GYRO_CONFIG0_FS_SEL_2000dps;

    switch (gyro_fsr_dps)
    {
    case 250:
        ret = GYRO_CONFIG0_FS_SEL_250dps;
        break;
    case 500:
        ret = GYRO_CONFIG0_FS_SEL_500dps;
        break;
    case 1000:
        ret = GYRO_CONFIG0_FS_SEL_1000dps;
        break;
    case 2000:
        ret = GYRO_CONFIG0_FS_SEL_2000dps;
        break;
    default:
        /* Unknown gyro FSR. Set to default 2000dps" */
        break;
    }
    return ret;
}

ACCEL_CONFIG0_ODR_t ICM42670P::accel_freq_to_param(uint16_t accel_freq_hz)
{
    ACCEL_CONFIG0_ODR_t ret = ACCEL_CONFIG0_ODR_100_HZ;

    switch (accel_freq_hz)
    {
    case 12:
        ret = ACCEL_CONFIG0_ODR_12_5_HZ;
        break;
    case 25:
        ret = ACCEL_CONFIG0_ODR_25_HZ;
        break;
    case 50:
        ret = ACCEL_CONFIG0_ODR_50_HZ;
        break;
    case 100:
        ret = ACCEL_CONFIG0_ODR_100_HZ;
        break;
    case 200:
        ret = ACCEL_CONFIG0_ODR_200_HZ;
        break;
    case 400:
        ret = ACCEL_CONFIG0_ODR_400_HZ;
        break;
    case 800:
        ret = ACCEL_CONFIG0_ODR_800_HZ;
        break;
    case 1600:
        ret = ACCEL_CONFIG0_ODR_1600_HZ;
        break;
    default:
        /* Unknown accel frequency. Set to default 100Hz */
        break;
    }
    return ret;
}

GYRO_CONFIG0_ODR_t ICM42670P::gyro_freq_to_param(uint16_t gyro_freq_hz)
{
    GYRO_CONFIG0_ODR_t ret = GYRO_CONFIG0_ODR_100_HZ;

    switch (gyro_freq_hz)
    {
    case 12:
        ret = GYRO_CONFIG0_ODR_12_5_HZ;
        break;
    case 25:
        ret = GYRO_CONFIG0_ODR_25_HZ;
        break;
    case 50:
        ret = GYRO_CONFIG0_ODR_50_HZ;
        break;
    case 100:
        ret = GYRO_CONFIG0_ODR_100_HZ;
        break;
    case 200:
        ret = GYRO_CONFIG0_ODR_200_HZ;
        break;
    case 400:
        ret = GYRO_CONFIG0_ODR_400_HZ;
        break;
    case 800:
        ret = GYRO_CONFIG0_ODR_800_HZ;
        break;
    case 1600:
        ret = GYRO_CONFIG0_ODR_1600_HZ;
        break;
    default:
        /* Unknown gyro ODR. Set to default 100Hz */
        break;
    }
    return ret;
}

static void event_cb(inv_imu_sensor_event_t *evt)
{
    //memcpy(event, evt, sizeof(inv_imu_sensor_event_t));
    memcpy(evt, event, sizeof(inv_imu_sensor_event_t));
}
